<?php

namespace OCA\GrueneGroupfolderManagement\Command;


use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;

use OC\Core\Command\Base;
use OC\Core\Command\InterruptedException;

use OCA\GrueneGroupfolderManagement\Db\GroupMapper;
use OCA\GrueneGroupfolderManagement\Db\GroupMemberMapper;
use OCA\GrueneGroupfolderManagement\Manager\GroupManager;
use OCA\GrueneGroupfolderManagement\Manager\SherpaManager;
use OCA\GrueneGroupfolderManagement\Manager\DirectoryManager;
use OCA\GrueneGroupfolderManagement\Manager\GroupFolderManager;
use OCA\GrueneGroupfolderManagement\Manager\DivisionManager;
use OCA\GrueneGroupfolderManagement\Constants\Constants as GrueneConst;
use OCA\GrueneGroupfolderManagement\Groups\GroupName;
use OCA\GrueneGroupfolderManagement\ApiClient;
use OCA\GrueneGroupfolderManagement\DTO\UpdateGroupDto;

use OCA\OrganizationFolders\Service\OrganizationFolderService;
use OCA\OrganizationFolders\Service\OrganizationFolderMemberService;
use OCA\OrganizationFolders\Service\ResourceService;
use OCA\OrganizationFolders\Service\ResourceMemberService;
use OCA\OrganizationFolders\Model\Principal;
use OCA\OrganizationFolders\Model\PrincipalFactory;
use OCA\OrganizationFolders\Enum\OrganizationFolderMemberPermissionLevel;
use OCA\OrganizationFolders\Enum\ResourceMemberPermissionLevel;
use OCA\OrganizationFolders\Enum\PrincipalType;
use OCA\OrganizationFolders\Errors\PrincipalAlreadyResourceMember;
use OCA\OrganizationFolders\Errors\PrincipalAlreadyOrganizationFolderMember;
use OCA\OrganizationFolders\Errors\PrincipalInvalid;
use OCA\OrganizationFolders\Errors\ResourceNameNotUnique;

class ConvertToOrganizationFolder extends Base {
	private ?OutputInterface $output = null;

	public function __construct(
		protected DirectoryManager $directoryManager,
		protected GroupMapper $groupMapper,
		protected GroupMemberMapper $groupMemberMapper,
		protected GroupManager $groupManager,
		protected GroupFolderManager $groupFolderManager,
		protected DivisionManager $divisionManager,
		protected OrganizationFolderService $organizationFolderService,
        protected OrganizationFolderMemberService $organizationFolderMemberService,
		protected ResourceService $resourceService,
		protected ResourceMemberService $resourceMemberService,
		protected PrincipalFactory $principalFactory,
		protected SherpaManager $sherpaManager,
		protected ApiClient $apiClient,
	) {
		parent::__construct();
	}

	protected function configure() {
		$this->setName('gruenegroupfoldermanagement:convert')
			->addArgument('division', InputArgument::OPTIONAL)
			->addOption("debug", null, InputOption::VALUE_NONE, "enable debug output")
			->addOption("all", null, InputOption::VALUE_NONE, "convert all instead of specific id")
			->addOption("offset", null, InputOption::VALUE_REQUIRED, "start --all migration after specific offset", 0)
			->setDescription('convert a specific folder to organization_folders');
		parent::configure();
	}

	private function debugLog(string $message) {
		$this->output->writeln("DEBUG: " . $message, OutputInterface::VERBOSITY_DEBUG);
	}

	private function errorLog(string $message) {
		$this->output->writeln("ERROR: " . $message);
	}

	private function infoLog(string $message) {
		$this->output->writeln("INFO: " . $message);
	}

    private function logProcess(string $message, callable $process) {
        $debugBuffer = [];

        $debugLog = function (string $message) use (&$debugBuffer) {
            $debugBuffer[] = $message;
        };

		$this->output->write("INFO: " . $message . "... ");

        try {
            $processOutput = $process($debugLog);
        } catch (\Throwable $e) {
            $error = $e;
        }

        if(isset($error)) {
            $this->output->writeln("error");
        } else {
            $this->output->writeln($processOutput ?? "done");
        }

        foreach($debugBuffer as $debugLine) {
            $this->output->writeln("DEBUG: " . $debugLine, OutputInterface::VERBOSITY_DEBUG);
        }

        $this->output->writeln("");

        if(isset($error)) {
            throw $error;
        }
	}

	protected function execute(InputInterface $input, OutputInterface $output): int {
		if($input->getOption("debug")) {
			$output->setVerbosity(OutputInterface::VERBOSITY_DEBUG);
		}
		$this->output = $output;

		$all = $input->getOption("all");

		if($all) {
			$offset = $input->getOption("offset");

			$divisions = $this->divisionManager->findAll();

			$divisionCount = count($divisions);

			$divisions = array_slice($divisions, $offset, preserve_keys: true);

			foreach($divisions as $index => $division) {
				try {
					$this->convertGroupfolderOfDivision($division->getId());
					$this->abortIfInterrupted();

					if($divisionCount > $index + 1) {
						$this->output->writeln("NOTE: To restart the migration from this point add --offset=" . $index + 1);
					}
					$this->output->writeln("");
				} catch (InterruptedException $e) {
					$this->output->writeln("Migration interrupted by user. To restart the migration from this point add --offset=" . $index + 1);
					break;
				} catch (\Throwable $e) {
					$this->output->writeln("MIGRATION FAILED! Details will be printed below. To retry the migration from this point add --offset=" . $index);
				}
			}

			return 0;
		} else {
			$divisionId = (int)$input->getArgument('division');
			$this->convertGroupfolderOfDivision($divisionId);

			return 0;
		}
	}

    /**
	 * @param string $mappingType
	 * @param string $mappingId
	 * @return Principal[]
	 */
    private function convertLegacyMappingToPrincipals(string $mappingType, string $mappingId, callable $debugLog): array {
        if($mappingType == "user") {
			return [$this->principalFactory->buildPrincipal(PrincipalType::USER, $mappingId)];

		} else if ($mappingType == "group") {
			if(str_starts_with($mappingId, GroupName::SHERPA_PREFIX)) {
                // sherpa role group

				[$prefix , $organizationKey, $slug] = explode("_", $mappingId, 3);

				//$debugLog("Parsed group name " . $mappingId .  " into " . json_encode(["organizationKey" => $organizationKey, "slug" => $slug]));

				$division = $this->divisionManager->find($organizationKey);

				return $this->convertLegacySherpaRoleGroupToPrincipals($division, $slug, $debugLog);
			} else if (preg_match('/^[0-9]+_[0-9]+$/', $mappingId)) {
                // sherpa role

                return [$this->principalFactory->buildPrincipal(
                    type: PrincipalType::ORGANIZATION_ROLE,
                    id: "sherpa:" . $mappingId
                )];
            } else if(str_starts_with($mappingId, GroupName::DIVISION_PREFIX)) {
                // sherpa organization member

                $organizationKey = explode("_", $mappingId, 3)[1];

                $division = $this->divisionManager->find($organizationKey);

                return [$this->principalFactory->buildPrincipal(
                    type: PrincipalType::ORGANIZATION_MEMBER,
                    id: "sherpa:" . $division->getShortKey()
                )];
            } else if(str_starts_with($mappingId, GroupName::GROUP_PREFIX)) {
                // skip, should have never been added as member

                return [];
            } else {
                // generic group

				return [$this->principalFactory->buildPrincipal(
                    type: PrincipalType::GROUP,
                    id: $mappingId,
                )];
			}
		} else {
            // should not be possible
            throw new \Exception("invalid mapping type, cannot convert");
        }
    }

	/**
	 * @param mixed $organization
	 * @param mixed $sherpaSlug
	 * @param mixed $level
	 * @return Principal[]
	 */
	private function convertLegacySherpaRoleGroupToPrincipals($division, $sherpaSlug, callable $debugLog): array {
		$group = $this->apiClient->groupBySlugAndLevel($division->getOrganization(), $sherpaSlug, $division->getLevel());

		$principals = [];

        if(isset($group)) {
            foreach(array_unique($group->getSherpaIds()) as $sherpaId) {
                $principals[] = $this->principalFactory->buildPrincipal(PrincipalType::ORGANIZATION_ROLE, "sherpa:" . $division->getShortKey() . "_" . $sherpaId);
            }
        } else {
            $debugLog("Sherpa Role Group " . $division->getOrganization()->getId() . "_" . $sherpaSlug . " could not be found in NB API, skipping");
        }

		return $principals;
	}

	protected function convertGroupMemberToResourceMembers(
        int $resourceId,
        string $mappingType,
        string $mappingId,
        string $type,
        callable $debugLog,
        ?string $ignoredRolePrincipalId = null,
    ) {
        $principals = $this->convertLegacyMappingToPrincipals($mappingType, $mappingId, $debugLog);

		$permissionLevel = ($type === "admin") ? ResourceMemberPermissionLevel::MANAGER : ResourceMemberPermissionLevel::MEMBER;

		$debugLog("Replacing group member " . json_encode([
			"mappingType" => $mappingType,
			"mappingId" => $mappingId
		]) . " with principals " . json_encode($principals) . " with permission level " . $permissionLevel->name);

		foreach($principals as $principal) {
			if($principal->isValid()) {
                if(!isset($ignoredRolePrincipalId) || $principal->getType() != PrincipalType::ORGANIZATION_ROLE || $principal->getId() != $ignoredRolePrincipalId) {
                    try {
                        $this->resourceMemberService->create(
                            resourceId: $resourceId,
                            principal: $principal,
                            permissionLevel: $permissionLevel,

                            skipPermssionsApply: true,
                        );
                    } catch (PrincipalAlreadyResourceMember $e) {
                        if($permissionLevel === ResourceMemberPermissionLevel::MANAGER) {
                            $existingMember = $this->resourceMemberService->findByPrincipal(
                                resourceId: $resourceId,
                                principal: $principal
                            );
                            
                            // if second occurance of principal has higher privileges update
                            if($existingMember->getPermissionLevel() === ResourceMemberPermissionLevel::MEMBER->value) {
                                $debugLog("Principal " . json_encode($principal) . " was already member of resource, but with lower permissions, promoting to MANAGER");

                                $this->resourceMemberService->update(
                                    id: $existingMember->getId(),
                                    permissionLevel: $permissionLevel,
                                );
                            } else {
                                $debugLog("Principal " . json_encode($principal) . " was already member of resource with same permissions, skipping");
                            }
                        }
                    }
                } else {
                    $debugLog("Principal " . json_encode($principal) . " was ignored");
                }
			} else {
				$debugLog("Principal " . json_encode($principal) . " is invalid (likely a user/group/role, that does not exist anymore), skipping");
			}
		}
	}

	protected function convertGroupfolderOfDivision(int $divisionId): void {
		$groupfolder = $this->groupFolderManager->getGroupFolderByOrganizationKey($divisionId);

		if(!$groupfolder) {
			$this->errorLog("divisionId is not valid, aborting");
			return;
		}

        $this->infoLog("Starting migration of groupfolder of division " . $divisionId);

		$this->debugLog("Converting this groupfolder: " . json_encode($groupfolder, JSON_PRETTY_PRINT));

		$division = $this->divisionManager->find($groupfolder['organization_key']);

		$this->debugLog("division shortKey " . $division->getShortKey());

        $organizationFolder = null;

        $this->logProcess("Setting up organization folder", function () use (&$organizationFolder, &$division, &$groupfolder) {
            $organizationFolder = $this->organizationFolderService->create(
                name: $groupfolder["mount_point"],
                quota: $groupfolder["quota"],
                organizationProvider: "sherpa",
                organizationId: $division->getShortKey(),

                existingGroupfolderId: $groupfolder["folder_id"],
            );
        });

        $managerPrincipal = null;

        $this->logProcess("Adding manager member to organization folder", function ($debugLog) use (&$organizationFolder, &$division, &$groupfolder, &$managerPrincipal) {
            // all technical wolke role groups only contain one role
            $managerPrincipal = $this->convertLegacySherpaRoleGroupToPrincipals(
                $division,
                $division->getType() . GroupName::TECHNICAL_WOLKE_SUFFIX,
                $debugLog
            )[0];

            try {
				if(!is_null($managerPrincipal)) {
					$this->organizationFolderMemberService->create(
						organizationFolder: $organizationFolder,
						permissionLevel: OrganizationFolderMemberPermissionLevel::MANAGER,
						principal: $managerPrincipal,
					);
					$debugLog("Added principal " . json_encode($managerPrincipal) . " as manager member");
				} else {
					$debugLog("manager principal for division " . $division->getShortKey() . " does not exist, skipping");
				}
            } catch (PrincipalAlreadyOrganizationFolderMember $e) {
                $debugLog("Principal " . json_encode($managerPrincipal) . " was already member of organization folder");
            } catch (PrincipalInvalid $e) {
                throw new \Exception("Could not add manager member to organization folder, because principal is invalid");
            }
        });

		$groups = $this->groupMapper->findMany(
			offset: 0,
			limit: 10000,
			divisions: [$groupfolder['organization_key']],
		);

		foreach ($groups as $group) {
			$this->logProcess("converting group " . $group->getName(), function ($debugLog) use (&$organizationFolder, &$group, &$managerPrincipal) {
				if(!$this->resourceService->isValidResourceName($group->getName())) {
					$fixedName = str_replace("*", "_", $group->getName());
					$fixedName = str_replace("/", "_", $fixedName);
					
					$debugLog("Group name is not compatible with organization folders, renaming to \"" . $fixedName . "\"");

					$group = $this->groupManager->update($group, new UpdateGroupDto(name: $fixedName));
				}

				try {
                    $resource = $this->resourceService->create(
                        type: "folder",
                        organizationFolderId: $organizationFolder->getId(),
                        name: $group->getName(),
                        parentResourceId: null,
                        active: $group->getActive(),
                        inheritManagers: true,
                        membersAclPermission: $group->getMemberAclPermission(),
                        managersAclPermission: $group->getAdminAclPermission(),
                        inheritedAclPermission: $group->getDivisionAclPermission(),
                        folderAlreadyExists: true,
                        skipPermssionsApply: true,
                    );
    
                    $groupMembers = $this->groupMemberMapper->findMany(
                        groupId: $group->getId(),
                        limit: 10000,
                    );
    
                    foreach($groupMembers as $groupMember) {
                        $this->convertGroupMemberToResourceMembers(
                            $resource->getId(),
                            $groupMember->getMappingType(),
                            $groupMember->getMappingId(),
                            $groupMember->getType(),
                            $debugLog,
                            $managerPrincipal?->getId(),
                        );
                    }
                } catch (ResourceNameNotUnique $e) {
                    return "has already been migrated, skipping";
                }
            });
		}

		if(in_array($division->getLevel(), [GrueneConst::COUNTY_CHAPTER_LEVEL, GrueneConst::LOCAL_GROUP_LEVEL])) {
			$this->logProcess("Converting special group " . GrueneConst::EXECUTIVE_BOARD, function ($debugLog) use (&$organizationFolder, &$division) {
                try {
                    $executiveBoardResource = $this->resourceService->create(
                        type: "folder",
                        organizationFolderId: $organizationFolder->getId(),
                        name: GrueneConst::EXECUTIVE_BOARD,
                        parentResourceId: null,
                        active: true,
                        inheritManagers: false,
                        membersAclPermission: 0,
                        managersAclPermission: 31,
                        inheritedAclPermission: 0,
                        folderAlreadyExists: true,
                    );

                    $this->convertGroupMemberToResourceMembers(
                        $executiveBoardResource->getId(),
                        "group",
                        GroupName::getExecutiveBoardGroupName($division),
                        "admin",
                        $debugLog
                    );
                } catch (ResourceNameNotUnique $e) {
                    return "has already been migrated, skipping";
                }
            });

			$this->logProcess("Converting special group " . GrueneConst::EXECUTIVE_BOARD_MANAGER, function ($debugLog) use (&$organizationFolder, &$division) {
                try {
                    $executiveBoardManagerResource = $this->resourceService->create(
                        type: "folder",
                        organizationFolderId: $organizationFolder->getId(),
                        name: GrueneConst::EXECUTIVE_BOARD_MANAGER,
                        parentResourceId: null,
                        active: true,
                        inheritManagers: false,
                        membersAclPermission: 0,
                        managersAclPermission: 31,
                        inheritedAclPermission: 0,
                        folderAlreadyExists: true,
                    );

                    $this->convertGroupMemberToResourceMembers(
                        $executiveBoardManagerResource->getId(),
                        "group",
                        GroupName::getExecutiveBoardGroupName($division),
                        "admin",
                        $debugLog
                    );
                    $this->convertGroupMemberToResourceMembers(
                        $executiveBoardManagerResource->getId(),
                        "group",
                        GroupName::getManagerGroupName($division),
                        "admin",
                        $debugLog
                    );
                } catch (ResourceNameNotUnique $e) {
                    return "has already been migrated, skipping";
                }
            });
		} else {
            $this->infoLog("No special groups to convert (not KV or OV)\n");
        }

		$this->logProcess("Applying ACL Rules to all the resources just created", function () use (&$organizationFolder) {
            $this->organizationFolderService->applyPermissions($organizationFolder);
        });

        $this->infoLog("Done migrating groupfolder of division " . $divisionId);
	}
}
